library ad_base;

import 'dart:async';
import 'package:flutter/foundation.dart';

/// 将tag转为对应广告SDK的广告单元的方法
typedef ConvertTagToUnitIdCallback = String Function(String tag);

abstract class ADInterface {
  Future<void> init();

  Future<void> setUserId(String userId);

  Future<AdResponse> loadAd(String unitId, AdType adType);

  Future<AdResponse> showAd(dynamic data, AdType adType);
}

class AdResponse {
  AdResponse({
    required this.status,
    this.message = '',
    this.data,
  });

  AdResponse.ok({String? message})
      : status = WatchAdStatus.ok,
        message = message ?? 'success',
        data = null;

  final WatchAdStatus status;
  final String message;
  final dynamic data;

  @override
  String toString() {
    return 'AdResponse{status: $status, message: $message}';
  }

  AdResponse copyWith({
    WatchAdStatus? status,
    String? message,
    dynamic data,
  }) {
    return AdResponse(
      status: status ?? this.status,
      message: message ?? this.message,
      data: data ?? this.data,
    );
  }

  void log() {
    if (kDebugMode) {
      print(toString());
    }
  }
}

/// 广告类型
enum AdType {
  // 激励视频
  rewardedVideo,
  // 插页
  interstitial,
  // banner
  banner,
}

/// 观看广告结果
enum WatchAdStatus {
  // 加载完广告
  loaded,
  // 成功观看完
  ok,
  // 用户放弃
  giveUp,
  // 无广告
  noAd,
  // 已在观看中
  watching,
  error,
}

class AdManager {
  /// 私有构造函数
  AdManager._internal();

  /// 保存单例
  static final AdManager _singleton = AdManager._internal();

  /// 工厂构造函数
  factory AdManager() => _singleton;

  late ADInterface adInterface;

  late ConvertTagToUnitIdCallback _convertTagToUnitId;

  final Map<String, Completer<AdResponse>> _loadMap = {};
  final List<String> _showingAd = [];

  /// 初始化
  /// [adInterface]广告SDK实现
  /// [convertTagToUnitId]将tag转为对应广告SDK的广告单元的方法
  /// [preloadAds]需要预加载的广告
  Future init(
    ADInterface adInterface,
    ConvertTagToUnitIdCallback convertTagToUnitId, {
    Map<AdType, List<String>>? preloadAds,
  }) async {
    try {
      this._convertTagToUnitId = convertTagToUnitId;
      await (this.adInterface = adInterface).init();

      preloadAds?.forEach((key, value) {
        for (final element in value) {
          _loadAd(element, key);
        }
      });
    } catch (e) {
      if (kDebugMode) {
        print(e.toString());
      }
    }
  }

  Future setUserId(String uid) async {
    await adInterface.setUserId(uid);
  }

  String _genKey(String tag, AdType adType) {
    return adType.toString() + '_' + tag;
  }

  Future<AdResponse> _loadAd(String tag, AdType adType) async {
    // 广告单元ID
    final unitId = _convertTagToUnitId(tag);
    // 生成key
    final key = _genKey(tag, adType);
    // 获取completer
    var completer = _loadMap[key];
    if (completer == null) {
      debugPrint('加载$adType广告');
      // 如果当前无key对应的任务
      completer = Completer<AdResponse>();
      _loadMap[key] = completer;
      // 加载广告
      adInterface.loadAd(unitId, adType).then(
        (value) {
          completer!.complete(value);
          if (value.status != WatchAdStatus.loaded) {
            _loadMap.remove(key);
          }
        },
        onError: (value) {
          completer!.complete(AdResponse(
            status: WatchAdStatus.error,
            message: value.toString(),
          ));
          _loadMap.remove(key);
        },
      );
    }
    // 返回当前key对应的任务
    return completer.future;
  }

  Future<AdResponse> showAd(
    String tag,
    AdType adType,
  ) async {
    final loadResult = await _loadAd(tag, adType);
    loadResult.log();
    if (loadResult.status == WatchAdStatus.loaded) {
      // 如果加载广告成功
      final key = _genKey(tag, adType);
      if (_showingAd.contains(key)) {
        return AdResponse(
          status: WatchAdStatus.watching,
          message: '该广告正在观看中',
        )..log();
      }
      // 观看广告
      _showingAd.add(key);
      final showResult = await adInterface.showAd(loadResult.data, adType);
      _showingAd.remove(key);
      // 清除对应key的任务
      _loadMap.remove(key);
      // 再次去加载广告
      _loadAd(tag, adType);
      if ([WatchAdStatus.ok, WatchAdStatus.giveUp].contains(showResult.status)) {
        return showResult..log();
      }
      return showResult.copyWith(status: WatchAdStatus.error)..log();
    } else if (loadResult.status == WatchAdStatus.ok) {
      return loadResult.copyWith(status: WatchAdStatus.error)..log();
    }
    return loadResult;
  }
}
